#include "coremoidapplet.h"
#include <QPainter>
#include <QFontMetrics>
#include <QSizeF>
#include <QTimer>
#include <QDir>
#include <QAction>
#include <QMenu>
#include <QApplication>
#include <QDBusMessage>
#include <QDBusConnection>

#include <KDialog>
#include <KColorScheme>
#include <KGlobalSettings>
#include <kiconloader.h>
#include <KSharedConfig>
#include <plasma/svg.h>
#include <plasma/theme.h>

#include "ui_coremoidConfig.h"

CoreMoidApplet::CoreMoidApplet ( QObject *parent, const QVariantList &args )
		: Plasma::Applet ( parent, args ),
		m_icon ( "document" ),
		m_size ( 100,50 )
{
	setAcceptsHoverEvents(true);
	m_config = NULL;
	m_dialog = NULL;
}

CoreMoidApplet::~CoreMoidApplet()
{
	if ( m_config )
		delete m_config;
	if ( m_dialog )
		delete m_dialog;
	m_timer->stop();
	delete m_timer;
}

void CoreMoidApplet::init()
{
	QDir dir;
	dir.setFilter(QDir::Dirs);
	dir.setPath("/sys/devices/system/cpu");
	m_cpu_count = dir.count();
	for(int i = 0; i < (int)dir.count(); i++)
		if ( dir[i] == "cpuidle" || dir[i] == "." || dir[i] == ".." )
			m_cpu_count--;
	kDebug() << "Number of CPU's " << m_cpu_count;
	for(int i = 0; i < m_cpu_count; i++)
		m_rects.push_back(QRect(5, 5 + (i * 30), 50, 20));
	m_stats.resize(m_cpu_count);
	m_stats_new.resize(m_cpu_count);

	readCpus(m_stats_new);
	readCpus(m_stats);
	readCurrentGovenor();

	m_timer = new QTimer(this);
	connect(m_timer, SIGNAL(timeout()), this, SLOT(doSomething()));
	m_timer->start(1000);

	m_performance = new QAction(SmallIcon("configure"), i18n("&Performance"), this);
	connect(m_performance, SIGNAL(triggered(bool)), this, SLOT(setPerformance()));
	m_performance->setCheckable(true);
	m_actions.append(m_performance);

	m_powersave = new QAction(SmallIcon("configure"), i18n("Power&save"), this);
	connect(m_powersave, SIGNAL(triggered(bool)), this, SLOT(setPowersave()));
	m_powersave->setCheckable(true);
	m_actions.append(m_powersave);

	m_ondemand = new QAction(SmallIcon("configure"), i18n("&Ondemand"), this);
	connect(m_ondemand, SIGNAL(triggered(bool)), this, SLOT(setOndemand()));
	m_ondemand->setCheckable(true);
	m_actions.append(m_ondemand);

	m_conservative = new QAction(SmallIcon("configure"), i18n("&Conservative"), this);
	connect(m_conservative, SIGNAL(triggered(bool)), this, SLOT(setConservative()));
	m_conservative->setCheckable(true);
	m_actions.append(m_conservative);

	updateGovenorSelection();

	m_font = QApplication::font();
	m_font.setWeight(QFont::Bold);
	
//	m_textColor = KColorScheme(QPalette::Active, KColorScheme::View, Plasma::Theme::self()->colors()).foreground().color();
//	m_boxColor = KColorScheme(QPalette::Active, KColorScheme::View, Plasma::Theme::self()->colors()).background().color();

	m_isHovered = false;
	m_boxAlpha = 128;
	m_boxHoverAlpha = 192;
	m_boxColor.setAlpha(m_boxAlpha);
	m_textColor = Plasma::Theme::self()->textColor();
	m_boxColor = Plasma::Theme::self()->backgroundColor();

	setAcceptsHoverEvents(true);
	setHasConfigurationInterface(true);

	KConfigGroup cg = config();
	m_orientation = cg.readEntry("orientation", 0);
	switch(m_orientation) {
	case 0:
		m_size.setWidth(100);
		m_size.setHeight(50);
		break;
	case 1:
		m_size.setWidth(60);
		m_size.setHeight(100);
		break;
	}

	m_show_governor = cg.readEntry("show_governor", true);
}

void CoreMoidApplet::setContentSize ( const QSizeF& size )
{
	m_size = size;
}

QSizeF CoreMoidApplet::contentSizeHint() const
{
	//If you want to change the size of your plasmoid, update m_size and call updateGeometry() afterwards.
	return m_size;
}

void CoreMoidApplet::paintInterface ( QPainter *p,
                                       const QStyleOptionGraphicsItem *option, const QRect &contentsRect )
{
	Q_UNUSED(option);

	p->setRenderHint ( QPainter::SmoothPixmapTransform );
	p->setRenderHint ( QPainter::Antialiasing );

	int text_height;
	if ( m_show_governor )
		text_height = 15;
	else
		text_height = 0;

	for(int i = 0; i < m_cpu_count; i++) {
		double user = m_stats_new[i].getUser() - m_stats[i].getUser();
		double system = m_stats_new[i].getSystem() - m_stats[i].getSystem();
		double nice = m_stats_new[i].getNice() - m_stats[i].getNice();
		double idle = m_stats_new[i].getIdle() - m_stats[i].getIdle();

		double usage = user + system + nice;
		usage = (double)100 / (double)(usage + idle) * usage;

		int height, width;
		switch(m_orientation) {
		case 0:
			height = (contentsRect.height() - text_height) / m_cpu_count;

			m_rects[i].setLeft(5);
			m_rects[i].setWidth(contentsRect.width() - 10);
			m_rects[i].setTop(height / 6 + i * height);
			m_rects[i].setHeight(height * 2 / 3);
			break;
		case 1:
			width = contentsRect.width() / m_cpu_count;

			m_rects[i].setLeft(width / 6 + i * width);
			m_rects[i].setWidth(width * 2 / 3);
			m_rects[i].setTop(5);
			m_rects[i].setHeight(contentsRect.height() - 10 - text_height);
			break;
		}
		paintBar(p, m_rects[i], (int)usage, m_orientation);
	}

	if ( m_show_governor ) {
		QRect rect_text(5, contentsRect.height() - text_height, contentsRect.width() - 10, text_height);
		paintText(p, rect_text, m_current_govenor);
	}

	if ( m_current_govenor == "ondemand" || m_current_govenor == "conservative" ) {
		int freq = getCurrentFreq();
		if (freq>=1000) {
			paintLabel(p, contentsRect, QString::number(freq/1000)+"."+QString::number((freq/100)%10)+ " GHz");
		} else {
			paintLabel(p, contentsRect, QString::number(freq) + " MHz");
		}
	}
}

/*!
    \fn CoreMoidApplet::doSomething()
 */
void CoreMoidApplet::doSomething()
{
	for(int i=0; i < m_cpu_count; i++)
		m_stats[i] = m_stats_new[i];
	readCpus(m_stats_new);
	readCurrentGovenor();
	updateGovenorSelection();
	update();
}

/*!
    \fn CoreMoidApplet::paintBar(QPainter *p)
 */
void CoreMoidApplet::paintBar(QPainter *p, const QRect &r, int perc, int orientation)
{
	p->save();
	p->setPen(Qt::gray);
	p->setBrush(Qt::gray);
	p->drawRect(r);
	p->setPen(Qt::green);
	p->setBrush(Qt::green);
	QRect r_tmp = r;
	if ( perc > 0 ) {
		switch(orientation) {
		default:
		case 0:
			r_tmp.setWidth(r_tmp.width() * perc / 100);
			break;
		case 1:
			int height = r_tmp.height() * perc / 100;
			r_tmp.setTop(r_tmp.top() + r_tmp.height() - height);
			r_tmp.setHeight(height);
			break;
		}
		p->drawRect(r_tmp);
	}
	p->restore();
}

void CoreMoidApplet::showConfigurationInterface()
{
	if ( m_dialog == NULL ) {
		m_dialog = new KDialog();
		m_dialog->setCaption(i18n("Configure CoreMoid"));
		m_dialog->setButtons( KDialog::Ok | KDialog::Cancel | KDialog::Apply );

		connect( m_dialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()) );
		connect( m_dialog, SIGNAL(okClicked()), this, SLOT(configAccepted()) );

		m_config = new Ui_coremoidConfig();
		m_config->setupUi(m_dialog->mainWidget());
	}

	switch(m_orientation) {
	case 0:
		m_config->rbVertical->setChecked(false);
		m_config->rbHorizontal->setChecked(true);
		break;
	case 1:
		m_config->rbVertical->setChecked(true);
		m_config->rbHorizontal->setChecked(false);
		break;
	}

	m_config->cbShowGovernor->setChecked(m_show_governor);

	m_dialog->show();
}

void CoreMoidApplet::configAccepted()
{
	KConfigGroup cg = config();
	int old_orientation = m_orientation;
	m_orientation = (m_config->rbVertical->isChecked() ? 1 : 0);
	if (old_orientation != m_orientation) {
		switch(m_orientation) {
		case 0:
			m_size.setWidth(100);
			m_size.setHeight(50);
			break;
		case 1:
			m_size.setWidth(50);
			m_size.setHeight(100);
			break;
		}
		updateGeometry();
	}
	cg.writeEntry("orientation", m_orientation);
	m_show_governor = m_config->cbShowGovernor->isChecked();
	cg.writeEntry("show_governor", m_show_governor);
	cg.sync();
	update();
}

void CoreMoidApplet::paintText(QPainter *p, const QRect &pos, const QString &text)
{
	p->save();
	p->setPen(Qt::gray);
	QString tmp = text;
	while (!textWillFit(p, pos, tmp))
		tmp.chop(1);
	p->drawText(pos, Qt::AlignCenter, tmp);
	p->restore();
}

void CoreMoidApplet::readCpus(QVector<CpuStat> &stats)
{
	QFile file_stats("/proc/stat", this);
	if ( !file_stats.open(QIODevice::ReadOnly | QIODevice::Text) ) {
		kDebug() << "File /proc/stat could not be opened";
		return;
	}

	QString all_files = file_stats.read(65535);
	QStringList lines = all_files.split('\n');
	for(int i = 0; i < lines.size(); i++) {
		if ( lines[i].startsWith("cpu") && lines[i][3] != ' ' ) {
			QStringList parts = lines[i].split(' ');
			int cpu_num = parts[0].mid(3).toInt();
			stats[cpu_num].setValues(parts[1].toInt(), parts[3].toInt(), parts[2].toInt(), parts[4].toInt());
		}
	}

	file_stats.close();
}

void CoreMoidApplet::readCurrentGovenor()
{
	QDBusMessage m = QDBusMessage::createMethodCall(
		"org.freedesktop.Hal",
		"/org/freedesktop/Hal/devices/computer",
		"org.freedesktop.Hal.Device.CPUFreq", 
		"GetCPUFreqGovernor");
	QDBusMessage response = QDBusConnection::systemBus().call(m);
	if ( response.type() != QDBusMessage::ReplyMessage ) {
		kDebug() << "DBus-Message Invalid Response received";
	}
	m_current_govenor = response.arguments().value(0).toString();
}

QList<QAction*> CoreMoidApplet::contextActions()
{
	//updateGovenorSelection();
	return m_actions;
}

void CoreMoidApplet::updateGovenorSelection()
{
	m_performance->setChecked(false);
	m_powersave->setChecked(false);
	m_ondemand->setChecked(false);
	m_conservative->setChecked(false);

	if ( m_current_govenor == "performance")
		m_performance->setChecked(true);

	else if ( m_current_govenor == "powersave")
		m_powersave->setChecked(true);

	else if ( m_current_govenor == "ondemand")
		m_ondemand->setChecked(true);

	else if ( m_current_govenor == "conservative")
		m_conservative->setChecked(true);
}

void CoreMoidApplet::setPerformance()
{
	setGovenor("performance");
	update();
}

void CoreMoidApplet::setPowersave()
{
	setGovenor("powersave");
	update();
}

void CoreMoidApplet::setOndemand()
{
	setGovenor("ondemand");
	update();
}

void CoreMoidApplet::setConservative()
{
	setGovenor("conservative");
	update();
}

void CoreMoidApplet::setGovenor(const QString &name)
{
	QDBusMessage m = QDBusMessage::createMethodCall(
		"org.freedesktop.Hal",
		"/org/freedesktop/Hal/devices/computer",
		"org.freedesktop.Hal.Device.CPUFreq",
		"SetCPUFreqGovernor");
	m << name;
	if ( !QDBusConnection::systemBus().send(m) ) {
		kDebug() << "DBus-Message can't be send";
	}
	m_current_govenor = name;
}

// This function ist adapted from battery applet
void CoreMoidApplet::paintLabel(QPainter *p, const QRect &contentsRect, const QString& labelText)
{
    // Store font size, we want to restore it shortly
    int original_font_size = m_font.pointSize();

    // Fonts smaller than smallestReadableFont don't make sense.
    m_font.setPointSize(qMax(KGlobalSettings::smallestReadableFont().pointSize(), m_font.pointSize()));
    QFontMetrics fm(m_font);
    int text_width = fm.width(labelText);

    // Longer texts get smaller fonts
    if (labelText.length() > 4) {
        if (original_font_size/1.5 < KGlobalSettings::smallestReadableFont().pointSize()) {
            m_font.setPointSize((int)(KGlobalSettings::smallestReadableFont().pointSize()));
        } else {
            m_font.setPointSize((int)(original_font_size/1.5));
        }
        fm = QFontMetrics(m_font);
        text_width = (int)(fm.width(labelText) * 1.2);
    } else {
        // Smaller texts get a wider box
        text_width = (int)(text_width * 1.4);
    }
    if (formFactor() == Plasma::Horizontal ||
        formFactor() == Plasma::Vertical) {
        m_font = KGlobalSettings::smallestReadableFont();
        m_font.setWeight(QFont::Bold);
        fm = QFontMetrics(m_font);
        text_width = (int)(fm.width(labelText)+8);
    } 
    p->setFont(m_font);

    // Let's find a good position for painting the background
    QRect text_rect = QRect(contentsRect.left()+(contentsRect.width()-text_width)/2,
                            (int)(contentsRect.top()+((contentsRect.height() - (int)fm.height())/2*0.9)),
                            text_width,
                            (int)(fm.height()*1.2));

    // Poor man's highlighting
    p->setPen(m_boxColor);
    if (m_isHovered) {
        m_boxColor.setAlpha(m_boxHoverAlpha);
    }
    p->setBrush(m_boxColor);

    // Find sensible proportions for the rounded corners
    float round_prop = text_rect.width() / text_rect.height();

    // Tweak the rounding edge a bit with the proportions of the textbox
    int round_radius = 35;
    p->drawRoundRect(text_rect, (int)(round_radius/round_prop), round_radius);

    p->setPen(m_textColor);
    p->drawText(text_rect, Qt::AlignCenter, labelText);

    // Reset font and box
    m_font.setPointSize(original_font_size);
    m_boxColor.setAlpha(m_boxAlpha);
}

int CoreMoidApplet::getCurrentFreq()
{
	QFile file_cur("/sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq", this);
	if ( !file_cur.open(QIODevice::ReadOnly | QIODevice::Text) ) {
		kDebug() << "File /sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq could not be opened";
		return -1;
	}

	QString ret = file_cur.read(256);
	ret.remove('\n');

	file_cur.close();

	return ret.toInt() / 1000;
}

void CoreMoidApplet::hoverEnterEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED( event );
    m_isHovered = true;
    update();
}

void CoreMoidApplet::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED( event );
    m_isHovered = false;
    update();
}

Qt::Orientations CoreMoidApplet::expandingDirections() const
{
	return 0;
}

void CoreMoidApplet::constraintsUpdated(Plasma::Constraints constraints)
{
	Q_UNUSED(constraints);
	updateGeometry();
}

bool CoreMoidApplet::textWillFit(QPainter *p, const QRect &pos, const QString &text) {
      QRect r = p->boundingRect ( 0,0,0,0, Qt::AlignCenter, text );
      return (r.width() < pos.width());
}

#include "coremoidapplet.moc"
